/*
 * This file is part of the Molecfit Pipeline
 * Copyright (C) 2001-2019 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
    #include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include <libgen.h>

#include <cpl.h>

#include "molecfit.h"

/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 *                 Typedefs: Structs and enum types
 */
/*----------------------------------------------------------------------------*/

typedef struct {

  cpl_boolean                      use_only_input_pri_ext;       /* If the user want to use only the SCIENCE_CALCTRANS/SCIENCE input FITS primary extension                                                            */
  int                              dflux_extension_data;         /* If use_only_input_pri_ext == CPL_TRUE, you can provide a extension as DFLUX (error flux)                                                           */
  int                              mask_extension_data;          /* If use_only_input_pri_ext == CPL_TRUE, you can provide a extension as MASK                                                                         */

  cpl_boolean                      suppress_extension;           /* Suppress arbitrary filename extension and change for numbers : [INPUT_NAME_1] => [OUTPUT_PRODUCT_1]_1 , ... [INPUT_NAME_N] => [OUTPUT_PRODUCT_N]_N */

  const char                       *mapping_correct;             /* Mapping extensions : inputs-telluric_corr,   i.e. {input_ext1-telluric_corr_ext,     ..., input_extN-telluric_corr_ext} int                        */
  cpl_table                        *mapping_correct_table;       /* Mapping telluric_corr extensions cpl_table. Contains 1 columns [TELLURIC_CORR_EXT]                                                                 */

  cpl_table 					   *fit_model_table;			 /* TELLURIC_CORR

  /* Telluric correction + column parameters */
  molecfit_fits                  *telluric_corr;                 /* cpl_table's in the TELLURIC_CORR (output of molecfit_calctrans)                                                                                    */
  const char                       *column_lambda;               /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_COLUMN_LAMBDA                                                                               */
  const char                       *column_flux;                 /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_COLUMN_FLUX                                                                                 */
  const char                       *column_dflux;                /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_COLUMN_DFLUX                                                                                */
  const char                       *column_mask;                 /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_COLUMN_MASK                                                                                 */

  cpl_boolean                      chip_extensions;              /* If the user want to combine the input FITS extensions  */
  const char                       *wlc_ref;                     /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_WAVELENGTH_FRAME                             													 */
  const char                       *wavelength_frame;            /* TELURIC_CORR molecfit parameter header : MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE                                                                                 */
  double 					  		observing_erv_rv_value; 	 /* values for observing_erv_rv.value */

  cpl_propertylist                 *pl;                          /* All input parameters to save in the output files                                                                                                   */

} molecfit_correct_parameter;

/*----------------------------------------------------------------------------*/
/**
 *                              Functions prototypes
 */
/*----------------------------------------------------------------------------*/

/* Fill the internal Molecfit configuration file */
static molecfit_correct_parameter * molecfit_correct_parameters(
    const cpl_frameset               *frameset,
    const cpl_parameterlist          *list);

/* Clean variables allocated in the recipe */
static void molecfit_correct_parameter_delete(
    molecfit_correct_parameter *parameters);

/*----------------------------------------------------------------------------*/
/**
 *                          Static variables
 */
/*----------------------------------------------------------------------------*/

#define RECIPE_NAME      MOLECFIT_CORRECT
#define CONTEXT          "molecfit."RECIPE_NAME

static char molecfit_correct_description[] =
    "This recipe reads the results from molecfit_calctrans and apply the telluric correction for scientific input data file.\n"
    "The input data "MOLECFIT_SCIENCE_CALCTRANS" or "MOLECFIT_SCIENCE" could be:\n"
    "  - BINTABLE  - Table  (single or multi-extension, optionally with some NULL extensions)\n"
    "  - IMAGE(1D) - Vector (single or multi-extension, optionally with some NULL extensions)\n"
    "  - IMAGE(2D) - Image  (single or multi-extension, optionally with some NULL extensions)\n"
    "It is not mandatory that all the extensions contain data.\n"
    "Molecfit will be run  on all the extensions that contain data.\n"
    "\n"
    "Input files:\n"
    "  - Mandatory, provide 1 to N files type "MOLECFIT_SCIENCE_CALCTRANS"/"MOLECFIT_SCIENCE".\n"
    "  - Mandatory, the telluric correction : It will apply in all of inputs. \n"
    "\n"
    "DO: Category                     Required   Explanation                               \n"
    "-------------                    --------   -----------                               \n"
    "  "MOLECFIT_SCIENCE_CALCTRANS" or "MOLECFIT_SCIENCE"     Y        The input data                    (Multi-extension FITS           format).\n"
    "  "MOLECFIT_TELLURIC_CORR"                   Telluric correction (transmission spectrum of the atmosphere) that needs to be applied (Multi-extension FITS IMAGE(1D) format).\n"
    "  "MOLECFIT_MAPPING_CORRECT"                  Y     List of the extensions to map from the telluric correction file "MOLECFIT_TELLURIC_CORR" to the extensions of the "MOLECFIT_SCIENCE_CALCTRANS" or "MOLECFIT_SCIENCE" input file.\n"

    "\n"
    "Notes:\n"
    "  - "MOLECFIT_MAPPING_CORRECT"  BINTABLE column  ["MOLECFIT_MAPPING_CORRECT_EXT"]\n"
    "\n"
    "\n"
    "Output files: (The output is 1 file for each input "MOLECFIT_SCIENCE_CALCTRANS"/"MOLECFIT_SCIENCE". The file match with the input dimensions)\n"
    "\n"
    "DO: Category                    Explanation\n"
    "-------------                   -----------\n"
    "  "MOLECFIT_CORRECT_CHIPS_COMBINED"        The input files chips combined in molecfit format (FITS BINTABLE    format           ).\n"
    "  "MOLECFIT_SCIENCE_TELLURIC_CORR"         The input  data corrected file                    (in the original  format           ).\n"
    "  "MOLECFIT_SPECTRUM_TELLURIC_CORR"        The output data corrected spectrum                (in the IMAGE(1D) format - Spectrum).\n"
    "\n";

/* Standard CPL recipe definition */
cpl_recipe_define(	molecfit_correct,
                  	MOLECFIT_BINARY_VERSION,
                  	"Jose A. Escartin",
                  	"https://support.eso.org/",
                  	"2019",
                  	"Read the results from molecfit_calctrans and apply the telluric correction for one or a set of scientific input data.",
                    molecfit_correct_description);


/*----------------------------------------------------------------------------*/
/**
 * @defgroup molecfit_correct  It runs molecfit to correct the input spectra.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 *                              Functions code
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @brief    Interpret the command line options and execute the data processing
 *
 * @param    frameset   the frames list
 * @param    parlist    the parameters list
 *
 * @return   CPL_ERROR_NONE if everything is OK or CPL_ERROR_CODE in other case
 *
 */
/*----------------------------------------------------------------------------*/
static int molecfit_correct(
    cpl_frameset             *frameset,
    const cpl_parameterlist  *parlist)
{
//cpl_frameset_dump(frameset,NULL);

  /* Get initial errorstate */
  cpl_error_ensure(frameset && parlist, CPL_ERROR_NULL_INPUT, return CPL_ERROR_NULL_INPUT, "NULL input : frameset and/or parlist");
  cpl_errorstate initial_errorstate = cpl_errorstate_get();

  /* Check frameset TAGS */
  cpl_msg_info(cpl_func, "Check frameset (SOF tags) ...");
  cpl_error_code err = molecfit_check_and_set_groups(frameset);

  /* Check mandatory TAGS/Parameters */
  if (!err) {

      /* SCIENCE_CALCTRANS/SCIENCE : Check only the first (at least need to contain one) */
      cpl_frame *input_frame_data = cpl_frameset_find(frameset, MOLECFIT_SCIENCE);
      if (!input_frame_data) {
          err = cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, "%s data not found in frameset!", MOLECFIT_SCIENCE);
      }

      /* TELLURIC_CORR */
      if (!err) {
          const cpl_frame *input_telluric_corr = cpl_frameset_find_const(frameset, MOLECFIT_TELLURIC_CORR);
          if (!input_telluric_corr) err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal INPUT TELLURIC_CORR!");
      }

      /* MAPPING_CORRECT */
      if (!err) {
          const cpl_frame *input_mapping_correct = cpl_frameset_find_const(frameset, MOLECFIT_MAPPING_CORRECT);
          if (!input_mapping_correct && (!strcmp(cpl_parameter_get_string(cpl_parameterlist_find_const(parlist, MOLECFIT_PARAMETER_MAPPING_CORRECT)), MF_PARAMETERS_NULL) ) ){
              err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal INPUT MAPPING_CORRECT!");
          }
      }

      /* FIT_MODEL*/
      /*if (!err) {
          const cpl_frame *fit_model = cpl_frameset_find_const(frameset, MOLECFIT_FIT_MODEL);
          cpl_msg_info(cpl_func, "FIT MODEL check done.  " );
          if (!fit_model && (!strcmp(cpl_parameter_get_string(cpl_parameterlist_find_const(parlist, MOLECFIT_PARAMETER_FIT_MODEL)), MF_PARAMETERS_NULL) ) ){
              err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal INPUT FIT_MODEL!");
          }
          cpl_msg_info(cpl_func, "got here without an issue. " );
      } */
  }
  
  
  /* TEST MAPPING CORRECT */
   if (!err) {
      cpl_error_code map_err=CPL_ERROR_NONE;
      map_err=molecfit_config_mappingchk(frameset,parlist,MOLECFIT_TELLURIC_CORR,
                                                          MOLECFIT_MAPPING_CORRECT,
														  "TELLURIC_CORR_EXT",
							  "MAPPING_CORRECT",
							  "SCIENCE");
      if (map_err) cpl_msg_error(cpl_func,"Mapping Errors detected cannot proceed further");
  }
     

  /* Recipe Parameters : Need scientific_header_primary */
  molecfit_correct_parameter *parameters = NULL;
  if (!err) {

      /* Get recipe parameters and update the molecfit default configuration */
      cpl_msg_info(cpl_func, "Load '%s' recipe parameters ...", MOLECFIT_CORRECT);
      parameters = molecfit_correct_parameters(frameset, parlist);

      if (!parameters) err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal INPUT PARAMETERS!");
      else             err = cpl_error_get_code();
  }


  /* Load TAG = FIT_MODEL */
  cpl_table *mapping_correct = NULL;
  if (!err) {

      if (parameters->mapping_correct_table) {
          mapping_correct = cpl_table_duplicate(parameters->mapping_correct_table);
      } else {
          cpl_msg_info (cpl_func, "Loading %s cpl_table", MOLECFIT_MAPPING_CORRECT);
          mapping_correct = molecfit_load_unique_table(frameset, MOLECFIT_MAPPING_CORRECT);
      }

      if (     !mapping_correct                                                           ) err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,      "Illegal INPUT MAPPING_CORRECT -> Not input!");
      else if (cpl_table_get_column_min(mapping_correct, MOLECFIT_MAPPING_CORRECT_EXT) < 0) err = cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "Illegal INPUT MAPPING_CORRECT -> Minimum column < 0!");
      else                                                                                  err = cpl_error_get_code();
  }

  /* Load TAG = MODEL */
  cpl_table *fit_model = NULL;
  if (!err) {

	  cpl_msg_info(cpl_func, " wlc_ref parameter:- %s ",parameters->wlc_ref);

	  if (!strcmp(parameters->wlc_ref,"MODEL")) {

		  if (parameters->fit_model_table) {
			  fit_model = cpl_table_duplicate(parameters->fit_model_table);
			  cpl_msg_info(cpl_func, "fit model table is found in the parameters. " );

		  } else {
			  cpl_msg_info (cpl_func, "Loading %s cpl_table", MOLECFIT_FIT_MODEL);
			  fit_model = molecfit_load_unique_table(frameset, MOLECFIT_FIT_MODEL);
			  cpl_msg_info(cpl_func, "fit model table is found and loaded." );
		  }


		  if (    (!fit_model)  &&  (!strcmp(parameters->wlc_ref,"MODEL"))                ) err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,      "Illegal INPUT FIT_MODEL -> Not input!");
		  else if (cpl_table_get_column_min(fit_model, MF_COL_MOD_LAMBDA) < 0) err = cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "Illegal INPUT FIT MODEL -> Minimum column < 0!");
		  else                                                                                  err = cpl_error_get_code();
	  }
  }



  /*** Correct SCIENCE_CALCTRANS/SCIENCE spectra ***/
  if (!err) {

      cpl_msg_info(cpl_func, "CORRECT input %s/%s files ", MOLECFIT_SCIENCE_CALCTRANS, MOLECFIT_SCIENCE);

      cpl_size n_frames = cpl_frameset_get_size(frameset);
      for (cpl_size frame = 0; frame < n_frames && !err; frame++) {

         /* Check all inputs in frameset to get only the SCIENCE_CALCTRANS/SCIENCE frames */
         cpl_frame  *frame_data = cpl_frameset_get_position(frameset, frame);
         const char *tag        = cpl_frame_get_tag(frame_data);
         if (   !strcmp(tag, MOLECFIT_SCIENCE_CALCTRANS)
             || !strcmp(tag, MOLECFIT_SCIENCE          ) ){

             /* Create a new independent cpl_frameset for this concrete input DATA frame */
             cpl_frameset *frameset_output = cpl_frameset_new();
             cpl_frameset_insert(frameset_output, cpl_frame_duplicate(frame_data));

             /* Load every frame in the input */
             const char    *filename = cpl_frame_get_filename(frame_data);
             molecfit_fits *data= molecfit_fits_load(filename, CPL_FALSE);
             //molecfit_fits *data     = molecfit_fits_load(filename, CPL_FALSE);


             /* Check DATA */
             if (!data) err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal INPUT %s cpl_frame!", tag);
             else       err = cpl_error_get_code();

             //PIPE-8525 - Part 1

             /* Execution extensions */
             cpl_size n_ext;
             if (     parameters->use_only_input_pri_ext) n_ext = 1;
             else if (parameters->chip_extensions       ) n_ext = data->v_ext[0].spectrum_data ? 1 : 2;
             else n_ext = data->n_ext;

             /* Convert to table for execute molecfit */
             if (!err) err = molecfit_data_convert_to_table( data,
                                                             parameters->chip_extensions,
                                                             parameters->use_only_input_pri_ext,
                                                             parameters->dflux_extension_data,
                                                             parameters->mask_extension_data,
                                                             parameters->column_lambda,
                                                             parameters->column_flux,
                                                             parameters->column_dflux,
                                                             parameters->column_mask);


             //PIPE-8525 - Part 2

             //REPLACE THE  FLUX/ lambda column with mlamnba of telluric_data
             //data->v_ext[ext].spectrum_data = cpl_strdup(MF_COL_MOD_LAMBDA); //

             if (!strcmp(parameters->wlc_ref,"MODEL")) {

            	 if (data->format == molecfit_fits_file_table){
            		 //parameters->column_lambda = cpl_strdup(MF_COL_MOD_LAMBDA);
            		 //data->v_ext[ext].spectrum_data = cpl_strdup(MF_COL_MOD_LAMBDA);
            	 }

            	 /*else if ((data->format == molecfit_fits_file_image_1D)|| molecfit_fits_file_image_2D ||
       				 molecfit_fits_file_image_3D){

            		 for (cpl_size ext = 0; ext < n_ext; ext++) {

            			 if (data->v_ext[ext].spectrum_data) {

            				 cpl_msg_info(cpl_func, "there's spectrum data here !!! ");
            				 cpl_table_select_all(data->v_ext[ext].spectrum_data);
            				 cpl_table *rangespec   = cpl_table_extract_selected(data->v_ext[ext].spectrum_data);
                   			 mf_convolution_rebin(data->v_ext[ext].spectrum_data, MF_COL_IN_LAMBDA, MF_COL_IN_FLUX,
                   					  rangespec, MF_COL_MOD_LAMBDA, MF_COL_IN_FLUX);
                   		 }
                   	  }
                  } */
             }

             //end PIPE-8525 - Part 2


             /* Save input DATA --> Combined (molecfit_spectrum in BINTABLE format) */
             if (parameters->chip_extensions) {

                 const char *filename_base = basename(data->filename);

                 char *out_combined = NULL;
                 if (parameters->suppress_extension) out_combined  = cpl_sprintf("%s_%lld.fits", MOLECFIT_CORRECT_CHIPS_COMBINED, frame        );
                 else                                out_combined  = cpl_sprintf("%s_%s",        MOLECFIT_CORRECT_CHIPS_COMBINED, filename_base);

                 cpl_size index_combined = data->v_ext[0].spectrum_data ? 0 : 1;
                 cpl_msg_info(cpl_func, "Save combined multi-extension molecfit_spectrum input FITS file ('%s') in ext=%lld ...", out_combined, index_combined);

                 err += molecfit_save(frameset, frameset, parlist, RECIPE_NAME, parameters->pl, MOLECFIT_CORRECT_CHIPS_COMBINED, out_combined);
                 if (!err) err = molecfit_save_mf_results(data->v_ext[index_combined].spectrum_head, out_combined, CPL_FALSE, NULL, data->v_ext[index_combined].spectrum_data, NULL);
             }

             /* Execution extensions */
             if (     parameters->use_only_input_pri_ext) n_ext = 1;
             else if (parameters->chip_extensions       ) n_ext = data->v_ext[0].spectrum_data ? 1 : 2;
             else n_ext = data->n_ext;

             /* Create output SPECTRUM SCIENCE FITS file */
//             molecfit_fits *spectrum_corr_data = NULL;
             if (!err) {

                 char *spectrum_filename = cpl_sprintf("SCIENCE_%s", basename(data->filename));
//                 spectrum_corr_data = molecfit_fits_create(spectrum_filename, molecfit_fits_file_image_1D, molecfit_fits_vector, data->n_ext);
                 cpl_free(spectrum_filename);

//                 if (spectrum_corr_data) err = cpl_error_get_code();
//                 if (!err)  err = cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_OUTPUT, "I cannot create the FITS output file %s!", MOLECFIT_SCIENCE_TELLURIC_CORR);


                 /* Create dummy extensions */
                 for (cpl_size ext = 0; ext < data->n_ext; ext++) {
                     if (data->v_ext[ext].spectrum_data) {
                         cpl_size size = -1;
                         if (     data->format == molecfit_fits_file_table   ) {
                             if (   cpl_propertylist_get_int(data->v_ext[ext].header, MOLECFIT_FITS_KEYWORD_NAXIS1) >  1
                                 && cpl_propertylist_get_int(data->v_ext[ext].header, MOLECFIT_FITS_KEYWORD_NAXIS2) == 1) {
                                 size = cpl_propertylist_get_int(data->v_ext[ext].header, MOLECFIT_FITS_KEYWORD_NAXIS1);
                             } else {
                                 size = cpl_table_get_nrow(data->v_ext[ext].table);
                             }
                         }
                         else if (data->format == molecfit_fits_file_image_1D) size = cpl_vector_get_size(   data->v_ext[ext].vector);
                         else if (data->format == molecfit_fits_file_image_2D) size = cpl_image_get_size_x(  data->v_ext[ext].image );
                         else if (data->format == molecfit_fits_file_image_3D) size = cpl_imagelist_get_size(data->v_ext[ext].cube  );

//                         spectrum_corr_data->v_ext[ext].vector = cpl_vector_new(size);
                     }
                     cpl_msg_info(cpl_func, "creating dummy extensions %d", (int)ext);


                 }
             }

             /* Loop over the data extensions */
             int null;
             for (cpl_size ext = 0; ext < n_ext && !err; ext++) {

                 /* Create spectrum header output */

 //                if (!err) spectrum_corr_data->v_ext[ext].header = cpl_propertylist_duplicate(data->v_ext[ext].spectrum_head);

                 /* Get data extension : Only manage extensions with data */
                 if (data->v_ext[ext].spectrum_data) {

                     /* Get TELLURIC_CORR */
                     cpl_size         index_telluric_corr_ext = cpl_table_get(mapping_correct, MOLECFIT_MAPPING_CORRECT_EXT, ext, &null);


                     const cpl_vector *telluric_corr          = parameters->telluric_corr->v_ext[index_telluric_corr_ext].vector;
                     cpl_vector_dump(parameters->telluric_corr->v_ext[0].vector , stdout);


                     cpl_msg_info(cpl_func, "loaded mapping correct   %d", (int)ext);
                     /* Check DATA with TELLURIC_CORR */
                     if (        !telluric_corr                                                                          ) err = cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "I cannot MATCH the TELLURIC_DATA extension with the %s extension for correct!", tag);
                     else if (   cpl_table_get_nrow(data->v_ext[ext].spectrum_data) <= 0
                              || cpl_table_get_nrow(data->v_ext[ext].spectrum_data) != cpl_vector_get_size(telluric_corr)) err = cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "I cannot MATCH the rows between the TELLURIC_DATA extension with the %s extension for correct!", tag);

                     if (!err) {

                    	 // PIPE-8582 - rebin data to mlambda column
                    	 /*if (parameters->wlc_ref =='MODEL')  {

                    		 cpl_msg_info(cpl_func, "going in to rebin stuff    %d", ext);

                    		 cpl_table_select_all(data->v_ext[ext].spectrum_data);
                    		 cpl_table *rangespec   = cpl_table_extract_selected(data->v_ext[ext].spectrum_data);

                    		 cpl_msg_info(cpl_func, "cpl_table rangespec created   %d", ext);
                    		 //cpl_table_dump(rangespec, 0, cpl_table_get_nrow(rangespec), NULL);

                    		 mf_convolution_rebin(data->v_ext[ext].spectrum_data, MF_COL_IN_LAMBDA, MF_COL_IN_FLUX, rangespec, MF_COL_MOD_LAMBDA, MF_COL_IN_FLUX);

                    		 cpl_msg_info(cpl_func, "rebinning worked   %d", ext);
                    		 cpl_table_delete(rangespec);
                    	 } */

                         /* Create input molecfit data */
                         cpl_msg_info(cpl_func, "Create input molecfit cpl_table data : %s [%s], ext=%lld", tag, data->filename, ext);
                         cpl_table *molecfit_data = mf_spectrum_create_input_to_correct( data->v_ext[ext].spectrum_data,
                                                                                         parameters->column_lambda,
                                                                                         parameters->column_flux,
                                                                                         parameters->column_dflux,
                                                                                         parameters->column_mask,
                                                                                         telluric_corr);

//                         cpl_table_dump_structure(molecfit_data, NULL);


                         if (molecfit_data) err = cpl_error_get_code();
                         else               err = cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "I cannot Create the input molecfit SPECTRUM_TABLE to correct!");


                         if (!strcmp(parameters->wlc_ref,"MODEL")  &&  (data->format == molecfit_fits_file_table)) {

                        	 err = cpl_table_erase_column( molecfit_data, MF_COL_IN_LAMBDA);
                        	 //cpl_type *dtype = pl_table_get_column_type(molecfit_data, MF_COL_IN_LAMBDA):
                        	 err = cpl_table_new_column(molecfit_data,MF_COL_IN_LAMBDA, CPL_TYPE_DOUBLE );
                        	 double *mlambda = cpl_table_get_data_double(fit_model, MF_COL_MOD_LAMBDA);
                        	 err = cpl_table_copy_data_double(molecfit_data, MF_COL_IN_LAMBDA, mlambda);

                         }


                         /* Apply correction to molecfit data */
                         if (!err) {
                             cpl_msg_info(cpl_func, "File : %s, ext=%lld : Apply correction", data->filename, ext);
                             err = mf_calctrans_correct_spectrum(molecfit_data, MF_PARAMETERS_TRANSMISSION_TRUE);

                         }

                         /* Correct for wavelength frame  */

                         if (!err) {
                        	 cpl_msg_info(cpl_func, "correcting back to wavelength frame ");
                        	 err = mf_spectrum_vacuum_to_air(parameters->wavelength_frame, parameters->observing_erv_rv_value, data->v_ext[ext].spectrum_data);
                         }


                         /* Create spectrum data output */
//                         if (!err) {
//
//                        	 cpl_msg_info(cpl_func, "creating spectrum data output  %d", (int)ext);
//                             /* Copy data */
//                             err = molecfit_data_update_vector(molecfit_data, parameters->chip_extensions, spectrum_corr_data, ext);
//
//                             /* Copy additional if chip combined */
//                             if (parameters->chip_extensions) {
//                                 for (cpl_size i = ext + 1; i < data->n_ext; i++) {
//                                     spectrum_corr_data->v_ext[i].header = cpl_propertylist_duplicate(data->v_ext[i].spectrum_head);
//                                 }
//                             }
//
//                         }

                         /* Update Input data with telluric correction */
                         if (!err) {
                             if (     data->format == molecfit_fits_file_table   ) err = molecfit_data_update_table( molecfit_data, parameters->chip_extensions, data, ext, parameters->column_flux);
                             else if (data->format == molecfit_fits_file_image_1D) err = molecfit_data_update_vector(molecfit_data, parameters->chip_extensions, data, ext);
                             else if (data->format == molecfit_fits_file_image_2D) err = molecfit_data_update_image( molecfit_data, parameters->chip_extensions, data, ext);
                             else if (data->format == molecfit_fits_file_image_3D) err = molecfit_data_update_cube(  molecfit_data, parameters->chip_extensions, data, ext);
                         }

                         /* Cleanup */
                         if (molecfit_data) cpl_table_delete(molecfit_data);
                     }
                 }
             }

             //cpl_propertylist_dump(data->v_ext[0].header, stdout);

             /*convert data back to reference frame */

             /* Save update OUT : SPECTRUM and DATA */
             if (!err) {

                 const char *filename_base = basename(data->filename);


                 /* Save input SPECTRUM extracted --> Corrected */
                 /*char *out_telluric_corr = NULL;
                 if (parameters->suppress_extension) out_telluric_corr  = cpl_sprintf("%s_%lld.fits", MOLECFIT_SPECTRUM_TELLURIC_CORR, frame        );
                 else                                out_telluric_corr  = cpl_sprintf("%s_%s",        MOLECFIT_SPECTRUM_TELLURIC_CORR, filename_base);

                 cpl_msg_info(cpl_func, "Writing on disk [%s] %s input FITS corrected on SPECTRUM format --> %s (PRO.CATG=%s)", data->filename, tag, out_telluric_corr, MOLECFIT_SPECTRUM_TELLURIC_CORR);
                 err = molecfit_fits_write(frameset, frameset_output, parlist, RECIPE_NAME, MOLECFIT_SPECTRUM_TELLURIC_CORR, spectrum_corr_data, out_telluric_corr);

                 cpl_free(out_telluric_corr); */


                 /* Save input SCIENCE --> Corrected */
                 if (!err) {

                     char *out_data = NULL;
                     if (parameters->suppress_extension) out_data = cpl_sprintf("%s_%lld.fits", MOLECFIT_SCIENCE_TELLURIC_CORR, frame        );
                     else                                out_data = cpl_sprintf("%s_%s",        MOLECFIT_SCIENCE_TELLURIC_CORR, filename_base);

                     cpl_msg_info(cpl_func, "Writing on disk [%s] %s input FITS corrected on the same format --> %s (PRO.CATG=%s)", data->filename, tag, out_data, MOLECFIT_SCIENCE_TELLURIC_CORR);
                     err = molecfit_fits_write(frameset, frameset_output, parlist, RECIPE_NAME, MOLECFIT_SCIENCE_TELLURIC_CORR, data, out_data);

                     cpl_free(out_data);
                 }
             }

             /* Cleanup */
             if (frameset_output   ) cpl_frameset_delete(frameset_output);
             if (data              ) molecfit_fits_delete(data);
//             if (spectrum_corr_data) molecfit_fits_delete(spectrum_corr_data);
         }
      }
  }

  /* Cleanup */
  if (parameters     ) molecfit_correct_parameter_delete( parameters     );
  if (mapping_correct) cpl_table_delete(                        mapping_correct);
  if (fit_model		 ) cpl_table_delete(                        fit_model);


  /* Check Recipe status and end */
  if (!err && cpl_errorstate_is_equal(initial_errorstate)) {
      cpl_msg_info(cpl_func,"Recipe successfully!");
  } else {
      /* Dump the error history */
      cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
      cpl_msg_error(cpl_func,"Recipe failed!, error(%d)=%s", err, cpl_error_get_message());
  }

  return err;
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
 * @brief Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param  self   parameterlist where you need put parameters
 *
 * @return cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code molecfit_correct_fill_parameterlist(
    cpl_parameterlist        *self)
{
  cpl_error_ensure(self, CPL_ERROR_NULL_INPUT, return CPL_ERROR_NULL_INPUT, "Null input");

  /* Add the different default parameters to the recipe */
  cpl_errorstate pre_state = cpl_errorstate_get();
  cpl_error_code e         = CPL_ERROR_NONE;

  /* Fill the parameters list */
  cpl_boolean    range     = CPL_TRUE;
  const void     *dummyMin = NULL;
  const void     *dummyMax = NULL;


  /* --MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI */
  cpl_boolean use_only_input_pri_ext = MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL, MF_STRLST_NULL,
						   &use_only_input_pri_ext,
                                                   MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI_DESC, CPL_FALSE);

  /* --MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA */
  int dflux_extension_data = MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
						   &dflux_extension_data,
                                                   MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA_DESC, CPL_FALSE);

  /* --MOLECFIT_PARAMETER_MASK_EXTENSION_DATA */
  int mask_extension_data = MOLECFIT_PARAMETER_MASK_EXTENSION_DATA_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_MASK_EXTENSION_DATA,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_INT, MF_STRLST_NULL,
						   &mask_extension_data,
                                                   MOLECFIT_PARAMETER_MASK_EXTENSION_DATA_DESC, CPL_FALSE);


  /* --MOLECFIT_PARAMETER_SUPPRESS_EXTENSION */
  cpl_boolean suppress_extension = MOLECFIT_PARAMETER_SUPPRESS_EXTENSION_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_SUPPRESS_EXTENSION,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL, MF_STRLST_NULL,
						   &suppress_extension,
                                                   MOLECFIT_PARAMETER_SUPPRESS_EXTENSION_DESC, CPL_FALSE);


  /* --MOLECFIT_PARAMETER_MAPPING_CORRECT */
  const char *mapping_correct = MOLECFIT_PARAMETER_MAPPING_CORRECT_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_MAPPING_CORRECT,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, MF_STRLST_NULL,
						   (const void *)mapping_correct,
                                                   MOLECFIT_PARAMETER_MAPPING_CORRECT_DESC, CPL_FALSE);

  /* --MOLECFIT_PARAMETER_MAPPING_CORRECT */
  /*const char *fit_model = MOLECFIT_PARAMETER_FIT_MODEL_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_FIT_MODEL,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, MF_STRLST_NULL,
						   (const void *)fit_model,
                                                   MOLECFIT_PARAMETER_FIT_MODEL_DESC, CPL_FALSE); */

  /* --MOLECFIT_PARAMETER_WLC_REF */
  const char *wlc_ref = MOLECFIT_PARAMETER_WLC_REF_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_WLC_REF,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING, MF_STRLST_NULL,
						  (const void *)wlc_ref,
                                                   MOLECFIT_PARAMETER_WLC_REF_DESC, CPL_FALSE);

  /* --MF_PARAMETERS_WAVELENGTH_FRAME */
  /*const char *wavelength_frame = MF_PARAMETERS_WAVELENGTH_FRAME_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MF_PARAMETERS_WAVELENGTH_FRAME,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_STRING,MF_STRLST_NULL,
                                                   (const void *)wavelength_frame,
                                                   MF_PARAMETERS_WAVELENGTH_FRAME_DESC, CPL_FALSE);
  */

  /* --MOLECFIT_PARAMETER_COMBINE_EXTENSIONS */
  cpl_boolean combine_extensions = MOLECFIT_PARAMETER_CHIP_EXTENSIONS_INIT;
  if (!e) e = molecfit_config_fill_parameter(RECIPE_NAME, self, MOLECFIT_PARAMETER_CHIP_EXTENSIONS,
                                                   !range, dummyMin, dummyMax, CPL_TYPE_BOOL, MF_STRLST_NULL,
						   &combine_extensions,
                                                   MOLECFIT_PARAMETER_CHIP_EXTENSIONS_DESC, CPL_FALSE);


  /* Check possible errors */
  if (!cpl_errorstate_is_equal(pre_state) || e != CPL_ERROR_NONE) {
      return cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                   "molecfit_correct_fill_parameterlist failed!");
  }

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief  Generate the internal configuration file for the recipe and check values
 *
 * @param  list   parameterlist with the parameters
 *
 * @return configuration file or NULL if exist an error
 *
 */
/*----------------------------------------------------------------------------*/
static molecfit_correct_parameter * molecfit_correct_parameters(
    const cpl_frameset              *frameset,
    const cpl_parameterlist         *list)
{
  /* Check input */
  cpl_error_ensure(frameset && list, CPL_ERROR_NULL_INPUT,
                   return NULL, "list input is NULL!");

  /* Get preState */
  cpl_errorstate preState = cpl_errorstate_get();
  const cpl_parameter *p;

  /* Create the configuration parameter */
  molecfit_correct_parameter *parameters = cpl_malloc(sizeof(molecfit_correct_parameter));
  parameters->mapping_correct_table            = NULL;
  parameters->fit_model_table	               = NULL;
  parameters->telluric_corr                    = NULL;
  parameters->pl                               = cpl_propertylist_new();


  /* Use only primary extension in the input FITS file ? */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI);
  parameters->use_only_input_pri_ext = cpl_parameter_get_bool(p);
  cpl_propertylist_update_bool(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_USE_ONLY_INPUT_PRI, parameters->use_only_input_pri_ext);

  /* If parameters->use_only_input_pri_ext == CPL_TRUE, you can provide a error flux extension */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA);
  parameters->dflux_extension_data = cpl_parameter_get_int(p);
  cpl_propertylist_update_int(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_DFLUX_EXTENSION_DATA, parameters->dflux_extension_data);

  /* If parameters->use_only_input_pri_ext == CPL_TRUE, you can provide a mask extension */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_MASK_EXTENSION_DATA);
  parameters->mask_extension_data = cpl_parameter_get_int(p);
  cpl_propertylist_update_int(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_MASK_EXTENSION_DATA, parameters->mask_extension_data);


  /* Mapping correct */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_MAPPING_CORRECT);
  parameters->mapping_correct = cpl_parameter_get_string(p);
  cpl_propertylist_update_string(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_MAPPING_CORRECT, parameters->mapping_correct);
  /* Create mapping correct cpl_table */
  if (strcmp(parameters->mapping_correct, MF_PARAMETERS_NULL)) {
      parameters->mapping_correct_table = molecfit_config_table_mapping(parameters->mapping_correct, MOLECFIT_MAPPING_CORRECT_EXT);
  }

  /* Suppress extension in the output recipe FITS files */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_SUPPRESS_EXTENSION);
  parameters->suppress_extension = cpl_parameter_get_bool(p);
  cpl_propertylist_update_bool(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_SUPPRESS_EXTENSION, parameters->suppress_extension);


  /* Load TAG = MOLECFIT_TELLURIC_CORR */
  cpl_msg_info(cpl_func, "Load %s from molecfit_calctrans(...)", MOLECFIT_TELLURIC_CORR);
  const cpl_frame *frm_telluric_corr = cpl_frameset_find_const(frameset, MOLECFIT_TELLURIC_CORR);

  if (cpl_frameset_find_const(frameset, MOLECFIT_TELLURIC_DATA)) {

  const cpl_frame *frm_telluric_data = cpl_frameset_find_const(frameset, MOLECFIT_TELLURIC_DATA);
  parameters->fit_model_table = molecfit_load_table(frm_telluric_data, 1);


  }

  if (frm_telluric_corr) {
      const char *filename = cpl_frame_get_filename(frm_telluric_corr);
      parameters->telluric_corr = molecfit_fits_load(filename, CPL_FALSE);
  }


  /* Combine extensions in the FITS file ? */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_CHIP_EXTENSIONS);
  parameters->chip_extensions = cpl_parameter_get_bool(p);
  cpl_propertylist_update_bool(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_CHIP_EXTENSIONS, parameters->chip_extensions);

  /* WLC_REF in the FITS file ?  */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_WLC_REF);
  parameters->wlc_ref = cpl_parameter_get_string(p);
  cpl_msg_info(cpl_func, "WLC_REF VALUE ->%s ", parameters->wlc_ref  );
  cpl_propertylist_update_string(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_WLC_REF, parameters->wlc_ref);

  cpl_msg_info(cpl_func, "WLC_REF VALUE  AFTER UPDATING->%s ", parameters->wlc_ref  );


  /* WAVELENGTH_FRAME in the FITS file ?  */
  /*p = cpl_parameterlist_find_const(list, MF_PARAMETERS_WAVELENGTH_FRAME);
  parameters->wavelength_frame = cpl_parameter_get_string(p);
  cpl_msg_info(cpl_func, "WAVELENGTH FRAME VALUE ->%s ", parameters->wavelength_frame  );
  cpl_propertylist_update_string(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_WAVELENGTH_FRAME, parameters->wavelength_frame);
  */

  p = cpl_parameterlist_find_const(list, MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE);
  if (p) {
	  parameters->observing_erv_rv_value = cpl_parameter_get_double(p);
      cpl_propertylist_update_double(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_OBSERVATORY_ERF_RV_VALUE, parameters->observing_erv_rv_value);
  }


  /* Load COLUMNS in TELLURIC_CORR HEADER */
  if (parameters->telluric_corr) {

      const char *column_str;

      column_str = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_LAMBDA;
      if (cpl_propertylist_has(parameters->telluric_corr->v_ext[0].header, column_str)) {
          parameters->column_lambda = cpl_strdup(cpl_propertylist_get_string(parameters->telluric_corr->v_ext[0].header, column_str));
      }

      column_str = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_FLUX;
      if (cpl_propertylist_has(parameters->telluric_corr->v_ext[0].header, column_str)) {
          parameters->column_flux   = cpl_strdup(cpl_propertylist_get_string(parameters->telluric_corr->v_ext[0].header, column_str));
      }

      column_str = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_DFLUX;
      if (cpl_propertylist_has(parameters->telluric_corr->v_ext[0].header, column_str)) {
          parameters->column_dflux  = cpl_strdup(cpl_propertylist_get_string(parameters->telluric_corr->v_ext[0].header, column_str));
      }

      column_str = MF_PARAMETERS_CONTEX_DEFAULT" "MF_PARAMETERS_COLUMN_MASK;
      if (cpl_propertylist_has(parameters->telluric_corr->v_ext[0].header, column_str)) {
          parameters->column_mask   = cpl_strdup(cpl_propertylist_get_string(parameters->telluric_corr->v_ext[0].header, column_str));
      }

      column_str = MF_PARAMETERS_CONTEX_DEFAULT" " MF_PARAMETERS_WAVELENGTH_FRAME;
      if (cpl_propertylist_has(parameters->telluric_corr->v_ext[0].header, column_str)) {
    	  parameters->wavelength_frame = cpl_strdup(cpl_propertylist_get_string(parameters->telluric_corr->v_ext[0].header, column_str));
      }

  }


  /* Combine extensions in the FITS file ? */
  p = cpl_parameterlist_find_const(list, MOLECFIT_PARAMETER_CHIP_EXTENSIONS);
  parameters->chip_extensions = cpl_parameter_get_bool(p);
  cpl_propertylist_update_bool(parameters->pl, MF_PARAMETERS_CONTEX_DEFAULT" "MOLECFIT_PARAMETER_CHIP_EXTENSIONS, parameters->chip_extensions);


  /* Check status */
  if (!cpl_errorstate_is_equal(preState)) {
      /* Configuration failed */
      molecfit_correct_parameter_delete(parameters);
      return NULL;
  }

  return parameters;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief    Deallocate the given parameter configuration object and its contents
 *
 * @param    parameters       The parameter configuration variable in the recipe.
 */
/*----------------------------------------------------------------------------*/
static void molecfit_correct_parameter_delete(
    molecfit_correct_parameter *parameters)
{
  if (parameters) {

      if (parameters->mapping_correct_table) cpl_table_delete(        parameters->mapping_correct_table);

      if (parameters->fit_model_table)		 cpl_table_delete(        parameters->fit_model_table);

      if (parameters->telluric_corr        ) molecfit_fits_delete(    parameters->telluric_corr        );

      if (parameters->pl                   ) cpl_propertylist_delete( parameters->pl                   );

      cpl_free(parameters);
  }
}
